# -*- coding: iso-8859-1 -*-
# pylint: disable-msg=W0131,E0213,F0203,E0201,W0232,R0921
# pylint-version = 0.7.0
#
# Copyright 2004-2005 Andr Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Testcases for svnmailer.processes

:var _tests: The test classes
:type _tests: ``list``
"""
__author__    = "Andr Malo"
__docformat__ = "restructuredtext en"
__all__       = ['getSuite']

from svnmailer_test import base
_tests = []

#### Test Bases                                     ####
########################################################
class TestBase(base.TestBase):
    """ Base class for processes tests """
    LOAD_MODULES = (('_proc', 'svnmailer.processes'),)

    def getInitialMocks(self):
        """ Returns the intially mocked modules """
        return [
            self._mock_os(),
        ]


#### Test Cases                                     ####
########################################################
class ProcessImplSubprocessTestCase(TestBase):
    """ svnmailer.processes.ProcessImpl (subprocess) """
    PREFIX = "ProcessImpl(subprocess)"

    def getInitialMocks(self):
        """ Mocks for subprocess implementation tests """
        mocks = super(ProcessImplSubprocessTestCase, self).getInitialMocks()
        mocks.extend([
            self._mock_svnmailer_util(),
            self._mock_subprocess(),
        ])
        return mocks

    def test00general(self):
        """ ProcessImpl(subprocess) is a private subprocess.Popen """
        expected = self.mock['subprocess'].Popen
        result = self._proc.ProcessImpl

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._proc),
            "Is public")

    def test01initial(self):
        """ .WINSHELL is properly filled initially """
        self.mock.result = [self._proc.ProcessImpl.WINSHELL]
        self.expect(False)

    def test02init(self):
        """ .__init__ initializes the attributes """
        import sys
        plat = sys.platform
        try:
            sys.platform = 'win32'
            inst = self._proc.ProcessImpl('foo', True, True, True, True)
            self.mock.result.extend([inst.tochild, inst.fromchild])
            inst = self._proc.ProcessImpl('foo', False, False, False, False)
            self.mock.result.extend([inst.tochild, inst.fromchild])
            sys.platform = 'other'
            inst = self._proc.ProcessImpl('foo', True, True, True, True)
            self.mock.result.extend([inst.tochild, inst.fromchild])
            inst = self._proc.ProcessImpl('foo', False, False, False, False)
            self.mock.result.extend([inst.tochild, inst.fromchild])
        finally:
            sys.platform = plat

        self.expect(
            'toLocale', 'foo', None, None, True,
            'subprocess.Popen', ('locale(foo, None)',), {'stdin': '*PIPE',
                'creationflags': 8, 'stderr': '*STDOUT', 'stdout': '*PIPE'},
            '*stdin', '*stdout',
            'toLocale', 'foo', None, None, True,
            'os.dup', 0,
            'os.dup', 1,
            'os.dup', 2,
            'subprocess.Popen', ('locale(foo, None)',), {'stdin': 3,
                'creationflags': 0, 'stderr': 5, 'stdout': 4},
            'os.close', 3,
            'os.close', 4,
            'os.close', 5,
            None, None,
            'subprocess.Popen', ('foo',), {'stdin': '*PIPE',
                'creationflags': 0, 'stderr': '*STDOUT', 'stdout': '*PIPE'},
            '*stdin', '*stdout',
            'os.dup', 0,
            'os.dup', 1,
            'os.dup', 2,
            'subprocess.Popen', ('foo',), {'stdin': 6, 'creationflags': 0,
                'stderr': 8, 'stdout': 7},
            'os.close', 6,
            'os.close', 7,
            'os.close', 8,
            None, None
        )

    def test03wait(self):
        """ .wait returns the exit status """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                pass
        inst = MyImpl()
        self.mock.result.append(inst.wait())

        self.expect('subprocess.Popen.wait', 0)

_tests.append(ProcessImplSubprocessTestCase)


class ProcessImplPopenTestCase(TestBase):
    """ svnmailer.processes.ProcessImpl (popen2.Popen3) """
    PREFIX = "ProcessImpl(popen)"

    def setUp(self):
        """ Loads the module independent from the platform """
        import sys
        plat = sys.platform
        try:
            sys.platform = 'other'
            super(ProcessImplPopenTestCase, self).setUp()
        finally:
            sys.platform = plat

    def getInitialMocks(self):
        """ Mocks for popen implementation tests """
        mocks = super(ProcessImplPopenTestCase, self).getInitialMocks()
        mocks.extend([
            ('subprocess', None), # ImportError
            self._mock_popen2(),
        ])
        return mocks

    def test00general(self):
        """ ProcessImpl(popen) is a private popen2.Popen3 """
        expected = self.mock['popen2'].Popen3
        result = self._proc.ProcessImpl

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._proc),
            "Is public")

    def test01general(self):
        """ ProcessImpl(popen) bails if selected on win32 """
        import sys
        def raiser():
            reload(self._proc)
        plat = sys.platform
        try:
            sys.platform = 'win32'
            self.failUnlessRaises(AssertionError, raiser)
        finally:
            sys.platform = plat

    def test02initial(self):
        """ .WINSHELL is properly filled initially """
        self.mock.result = [self._proc.ProcessImpl.WINSHELL]
        self.expect(False)

    def test03init(self):
        """ .__init__ forks and runs the child """
        self.mock.os.forkpids = [0]
        self._proc.ProcessImpl('foo', True, True, True, True)
        self._proc.ProcessImpl('foo', False, False, False, False)

        self.expect(
            'popen2._cleanup',
            'os.pipe',
            'os.pipe',
            'os.fork',
            'os.dup2', 3, 0,
            'os.dup2', 6, 1,
            'os.dup2', 6, 2,
            'popen2.Popen3._run_child', 'foo',

            'popen2._cleanup',
            'os.fork',
            'popen2.Popen3._run_child', 'foo'
        )

    def test04init(self):
        """ .__init__ forks and runs the parent """
        self.mock.os.forkpids = [1, 2]
        inst1 = self._proc.ProcessImpl('foo', True, True, True, True)
        self.mock.result.extend([inst1.tochild, inst1.fromchild])
        inst2 = self._proc.ProcessImpl('foo', False, False, False, False)
        self.mock.result.extend([inst2.tochild, inst2.fromchild])
        self.mock.result.extend(self.mock['popen2']._active)

        self.expect(
            'popen2._cleanup',
            'os.pipe',
            'os.pipe',
            'os.fork',
            'os.close', 3,
            'os.fdopen', 4, 'wb', -1,
            'os.close', 6,
            'os.fdopen', 5, 'rb', -1,
            '*file4', '*file5',

            'popen2._cleanup',
            'os.fork',
            None, None, inst1, inst2
        )

    def test05wait(self):
        """ .wait returns the exit status """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                pass
        inst = MyImpl()
        self.mock.result.append(inst.wait())

        self.expect('popen2.Popen3.wait', 0)

_tests.append(ProcessImplPopenTestCase)


class ProcessImplWin32apiTestCase(TestBase):
    """ svnmailer.processes.ProcessImpl (win32api) """
    PREFIX = "ProcessImpl(win32api)"

    def setUp(self):
        """ Loads the module independent from the platform """
        import sys
        plat = sys.platform
        try:
            sys.platform = 'win32'
            super(ProcessImplWin32apiTestCase, self).setUp()
        finally:
            sys.platform = plat

    def getInitialMocks(self):
        """ Mocks for win32api implementation tests """
        mocks = super(ProcessImplWin32apiTestCase, self).getInitialMocks()
        mocks.extend([
            ('subprocess', None), # ImportError
            ('popen2', {}),       # AttributeError
            self._mock_win32api(),
            self._mock_win32con(),
            self._mock_win32pipe(),
            self._mock_win32process(),
            self._mock_win32event(),
            self._mock_win32file(),
            self._mock_msvcrt(),
        ])
        return mocks

    def test00general(self):
        """ ProcessImpl(win32api) is a private object """
        expected = object
        result = self._proc.ProcessImpl

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._proc),
            "Is public")

    def test01general(self):
        """ ProcessImpl(win32api) bails if selected on non-win32 """
        import sys
        def raiser():
            reload(self._proc)
        plat = sys.platform
        try:
            sys.platform = 'other'
            self.failUnlessRaises(AssertionError, raiser)
        finally:
            sys.platform = plat

    def test02initial(self):
        """ .WINSHELL is properly filled initially """
        self.mock.result = [self._proc.ProcessImpl.WINSHELL]
        self.expect(False)

    def test03init(self):
        """ .__init__ initializes attributes """
        class MyImpl(self._proc.ProcessImpl):
            def _dup(this, handle):
                self.mock.result.extend(['_dup', handle.value])
                return handle
        inst = MyImpl('foo', True, True, True, True)
        self.mock.result.extend([
            inst.tochild, inst.fromchild, inst._handle
        ])
        inst = MyImpl('foo', False, False, False, False)
        self.mock.result.extend([
            inst.tochild, inst.fromchild, inst._handle.value
        ])

        self.expect(
            'win32pipe.CreatePipe', None, 0,
            'Handle.Detach', 4,
            'msvcrt.open_osfhandle', 4, 0,
            'os.fdopen', 4, 'wb', -1,
            '_dup', 3,
            'win32pipe.CreatePipe', None, 0,
            'Handle.Detach', 5,
            'msvcrt.open_osfhandle', 5, 0,
            'os.fdopen', 5, 'rb', -1,
            '_dup', 6,
            '_dup', 6,
            'win32process.CreateProcess', None, 'foo', None, None, 1, 8,
                None, None, 1, 3, 6, 6,
            'Handle.Close', 100,
            'Handle.Close', 200,
            'win32file.CloseHandle', 3,
            'win32file.CloseHandle', 6,
            'win32file.CloseHandle', 6,
            '*file4', '*file5', None,

            'win32api.GetStdHandle', -1,
            '_dup', 0,
            'win32api.GetStdHandle', -2,
            '_dup', 1,
            'win32api.GetStdHandle', -3,
            '_dup', 2,
            'win32process.CreateProcess', None, 'foo', None, None, 1, 0,
                None, None, 1, 0, 1, 2,
            'Handle.Close', 200,
            'win32file.CloseHandle', 0,
            'win32file.CloseHandle', 1,
            'win32file.CloseHandle', 2,
            None, None, 100
        )

    def test04wait(self):
        """ .wait returns 0 on detached process """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                this._handle = None
        self.mock.result.append(MyImpl().wait())

        self.expect(0)

    def test05wait(self):
        """ .wait on normal process returns the exit status """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                this._handle = self._win32_mockhelper()['Handle'](100)
        self.mock.result.append(MyImpl().wait())

        self.expect(
            'win32event.WaitForSingleObject', 100, -1,
            'win32process.GetExitCodeProcess', 100,
            'Handle.Close', 100,
            31415
        )

    def test06dup(self):
        """ ._dup duplicates the handle """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                pass
        self.mock.result.append(MyImpl()._dup(2))

        self.expect(
            'win32api.GetCurrentProcess',
            'win32api.DuplicateHandle', -1, 2, -1, 0, 1, 3,
            -2
        )

_tests.append(ProcessImplWin32apiTestCase)


class ProcessImplDumbTestCase(TestBase):
    """ svnmailer.processes.ProcessImpl (dumb) """
    PREFIX = "ProcessImpl(dumb)"

    def getInitialMocks(self):
        """ Mocks for dumb implementation tests """
        mocks = super(ProcessImplDumbTestCase, self).getInitialMocks()
        mocks.extend([
            ('subprocess', None), # ImportError
            ('popen2', {}),       # AttributeError
            ('win32api', None),   # ImportError
        ])
        return mocks

    def test00general(self):
        """ ProcessImpl(dumb) is a private object """
        expected = object
        result = self._proc.ProcessImpl

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._proc),
            "Is public")

    def test01initial(self):
        """ .WINSHELL is properly filled initially """
        self.mock.result = [self._proc.ProcessImpl.WINSHELL]
        self.expect(True)

    def test02init(self):
        """ .__init__ raises NotImplementedError """
        self.failUnlessRaises(NotImplementedError,
            self._proc.ProcessImpl, 'foo', False, True, True, False)
        self.failUnlessRaises(NotImplementedError,
            self._proc.ProcessImpl, 'foo', True, False, True, False)
        self.failUnlessRaises(NotImplementedError,
            self._proc.ProcessImpl, 'foo', True, True, True, True)

    def test03init(self):
        """ .__init__ initializes attributes """
        inst = self._proc.ProcessImpl('foo', True, True, True, False)
        self.mock.result.extend([inst.tochild, inst.fromchild])
        inst = self._proc.ProcessImpl('foo', True, True, False, False)
        self.mock.result.extend([inst.tochild, inst.fromchild])

        self.expect(
            'os.popen4', 'foo', 'b', -1,
            '*file3', '*file4',
            'os.popen2', 'foo', 'b', -1,
            '*file5', '*file6'
        )

    def test04wait(self):
        """ .wait returns 0 """
        class MyImpl(self._proc.ProcessImpl):
            def __init__(this):
                pass
        self.mock.result.append(MyImpl().wait())

        self.expect(0)

_tests.append(ProcessImplDumbTestCase)


class ProcessTestCase(TestBase):
    """ svnmailer.processes.Process """
    PREFIX = "Process"

    def getInitialMocks(self):
        """ Mocks for Process tests """
        mocks = super(ProcessTestCase, self).getInitialMocks()
        mocks.extend([
            ('subprocess', None), # ImportError
            ('popen2', {}),       # AttributeError
            ('win32api', None),   # ImportError
        ])
        return mocks

    def test00general(self):
        """ Process is a public object """
        expected = object
        result = self._proc.Process

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._proc),
            "Is not public")

    def test01initial(self):
        """ .IMPLEMENTATION is properly filled initially """
        import sys
        plat = sys.platform
        try:
            self.addMock(*self._mock_subprocess())
            reload(self._proc)
            self.mock.result.append(self._proc.Process.IMPLEMENTATION)
    
            self.addMock('subprocess', None)
            self.addMock(*self._mock_popen2())
            sys.platform = 'other'
            reload(self._proc)
            self.mock.result.append(self._proc.Process.IMPLEMENTATION)

            self.addMock('popen2', {})
            self.addMock(*self._mock_win32api())
            self.addMock(*self._mock_win32con())
            self.addMock(*self._mock_win32pipe())
            self.addMock(*self._mock_win32process())
            self.addMock(*self._mock_win32event())
            self.addMock(*self._mock_win32file())
            self.addMock(*self._mock_msvcrt())
            sys.platform = 'win32'
            reload(self._proc)
            self.mock.result.append(self._proc.Process.IMPLEMENTATION)

            self.addMock('win32api', None)
            reload(self._proc)
            self.mock.result.append(self._proc.Process.IMPLEMENTATION)
        finally:
            sys.platform = plat

        self.expect('subprocess', 'popen', 'win32api', 'dumb')

    def test02initial(self):
        """ ._IMPLEMENTATION_CLASS is properly filled initially """
        res = self._proc.Process._IMPLEMENTATION_CLASS
        self.failUnless(issubclass(res, self._proc.ProcessImpl),
            "No ProcessImpl")

    def test03initial(self):
        """ ._SLASHRE and ._TOKENRE are properly filled initially """
        import sys
        self.addMock(*self._mock_re())
        plat = sys.platform
        try:
            sys.platform = "other"
            reload(self._proc)
            self.failIf(hasattr(self._proc.Process, '_SLASHRE'),
                "regex exists")
            self.failIf(hasattr(self._proc.Process, '_TOKENRE'),
                "regex exists")

            sys.platform = "win32"
            reload(self._proc)
            self.mock.result.extend([
                self._proc.Process._SLASHRE._regex.pattern,
                self._proc.Process._TOKENRE._regex.pattern
            ])
        finally:
            sys.platform = plat

        self.expect(r'(\\+)("|$)', r'["\s]')

    def test04init(self):
        """ .__init__ initializes attributes """
        class MyImpl(object):
            WINSHELL = 'flopp'
            def __init__(this, *args, **kwargs):
                this.__mocked__ = '*MyImpl'
                self.mock.result.extend(['MyImpl', args, kwargs])
                this.tochild = '*tochild'
                this.fromchild = '*fromchild'
        class MyProcess(self._proc.Process):
            _IMPLEMENTATION_CLASS = MyImpl
            def _escapecmd(this, command, winshell = None):
                self.mock.result.extend(['_escapecmd', command, winshell])
                return "escaped(%s)" % command
        inst = MyProcess('foo')
        self.mock.result.extend([
            inst.tochild, inst.fromchild, inst._impl.__mocked__
        ])
        inst = MyProcess('bar', 1, 2, 3, 4)
        self.mock.result.extend([
            inst.tochild, inst.fromchild, inst._impl.__mocked__
        ])

        self.expect(
            '_escapecmd', 'foo', 'flopp',
            'MyImpl', ('escaped(foo)', False, False, False, False), {},
            '*tochild', '*fromchild', '*MyImpl',
            '_escapecmd', 'bar', 'flopp',
            'MyImpl', ('escaped(bar)', 1, 2, 3, 4), {},
            '*tochild', '*fromchild', '*MyImpl'
        )

    def test05pipe2(self):
        """ .pipe2 returns a properly initialized instance """
        class MyProcess(self._proc.Process):
            def __init__(this, cmd, pstdin = False, pstdout = False,
                         pstderr = False, detach = False):
                self.mock.result.extend(['MyProcess',
                    cmd, pstdin, pstdout, pstderr, detach
                ])
        inst = MyProcess.pipe2('foo')

        self.failUnless(isinstance(inst, self._proc.Process), 'No Process')
        self.expect('MyProcess', 'foo', True, True, False, False)

    def test05pipe4(self):
        """ .pipe4 returns a properly initialized instance """
        class MyProcess(self._proc.Process):
            def __init__(this, cmd, pstdin = False, pstdout = False,
                         pstderr = False, detach = False):
                self.mock.result.extend(['MyProcess',
                    cmd, pstdin, pstdout, pstderr, detach
                ])
        inst = MyProcess.pipe4('bar')

        self.failUnless(isinstance(inst, self._proc.Process), 'No Process')
        self.expect('MyProcess', 'bar', True, True, True, False)

    def test06detach(self):
        """ .detach returns a properly initialized instance """
        class MyProcess(self._proc.Process):
            def __init__(this, cmd, pstdin = False, pstdout = False,
                         pstderr = False, detach = False):
                self.mock.result.extend(['MyProcess',
                    cmd, pstdin, pstdout, pstderr, detach
                ])
        inst = MyProcess.detach('baz')

        self.failUnless(isinstance(inst, self._proc.Process), 'No Process')
        self.expect('MyProcess', 'baz', False, False, False, True)

    def test07detach(self):
        """ .detach passes NotImplementedError """
        class MyProcess(self._proc.Process):
            def __init__(this, cmd, pstdin = False, pstdout = False,
                         pstderr = False, detach = False):
                raise NotImplementedError()

        self.failUnlessRaises(NotImplementedError, MyProcess.detach, 'zonk')

    def test08wait(self):
        """ .wait returns the implementation wait return """
        class MyImpl(object):
            def wait(self):
                return 'blah'
        class MyProcess(self._proc.Process):
            def __init__(this):
                this._impl = MyImpl()
        self.mock.result.append(MyProcess().wait())

        self.expect('blah')

    def _doescape(self, platform, winshell, *tests):
        """ Do the escapecmd tests """
        import sys
        plat = sys.platform
        try:
            sys.platform = platform
            reload(self._proc)
            class MyProcess(self._proc.Process):
                def __init__(this):
                    pass
            inst = MyProcess()
            for cmd in tests:
                self.mock.result.append(
                    inst._escapecmd(cmd, winshell = winshell)
                )
        finally:
            sys.platform = plat

    def test09escapecmd(self):
        """ ._escapecmd returns the command itself on non-win32 """
        self._doescape('other', False, ['foo', 'bar'])
        self._doescape('other', True, ['baz', 'zonk'])

        self.expect(['foo', 'bar'], ['baz', 'zonk'])

    def test10escapecmd(self):
        """ ._escapecmd returns a suitable string for win32 shell """
        tests = (
            ['foo'],
            ['perl', '-le', 'print "42"'],
            [r'C:\Program Files\tools\diff.exe', '-u', '-L', 'file1',
                '-L', 'file2', r'C:\Temp\tmpfile1', r'C:\Temp\tmpfile2'],
        )
        for test in tests:
            self._doescape('win32', True, test)

        self.expect(
            '""foo""',
            '""perl" "-le" "print "^""42"^""""',
           r'""C:\Program Files\tools\diff.exe" "-u" "-L" "file1" '
                r'"-L" "file2" "C:\Temp\tmpfile1" "C:\Temp\tmpfile2""',
        )

    def test11escapecmd(self):
        """ ._escapecmd returns a suitable string for win32 non-shell """
        tests = (
            ['foo'],
            ['perl', '-le', 'print "42"'],
            [r'C:\Program Files\tools\diff.exe', '-u', '-L', 'file1',
                '-L', 'file2', r'C:\Temp\tmpfile1', r'C:\Temp\tmpfile2'],
        )
        for test in tests:
            self._doescape('win32', False, test)

        self.expect(
            'foo',
           r'perl -le "print \"42\""',
           r'"C:\Program Files\tools\diff.exe" -u -L file1 '
                r'-L file2 C:\Temp\tmpfile1 C:\Temp\tmpfile2',
        )

_tests.append(ProcessTestCase)


#### Glue code                                      ####
########################################################
__all__.extend(_tests)

def getSuite():
    """ Returns the test suite for this module """
    return [_test.__name__ for _test in _tests]
